"""This function removes the background from spots intensity.

Given a folder with analyzed data, the algorithm evaluates the
background value of each spot in each frame (using cages) and
subtracte it from the intensity of the spot. The background value
is multiplied by the volume of the spots of course. results
written in a .xls file.
"""

import datetime
import multiprocessing
import numpy as np
import xlwt

import SpotsBackgroundEstimation


class WriteSptsIntsDividedByBkg:
    """Principal class, organizes the multiprocessing"""
    def __init__(self, foldername, green4D, spots_3D, spots_tracked_3D):

        spts_id  =  np.unique(spots_tracked_3D)[1:].astype(np.int)                      # list of spots id
        steps    =  spots_tracked_3D.shape[0]                                           # number of time steps

        cpu_ow     =  multiprocessing.cpu_count()
        jobs_args  =  []
        t_chops    =  np.linspace(0, steps, cpu_ow).astype(np.int)
        for k in range(cpu_ow - 1):                                                     # organize input data in lists for multiprocessing
            jobs_args.append([green4D[t_chops[k]:t_chops[k + 1], :, :, :], spots_3D.spots_ints[t_chops[k]:t_chops[k + 1], :, :], spots_3D.spots_vol[t_chops[k]:t_chops[k + 1], :, :], spots_tracked_3D[t_chops[k]:t_chops[k + 1], :, :], spts_id])

        pool     =  multiprocessing.Pool()
        results  =  pool.map(WriteSptsIntsDividedByBkgUtility, jobs_args)
        pool.close()

        spts_int_bybkg     =  results[0].spts_int_bybkg                             # read and concatenate multiprocessing results
        av_spts_int_bybkg  =  results[0].av_spts_int_bybkg
        bkg                =  results[0].bkg
        bkg_variab         =  results[0].bkg_variab
        for j in range(1, cpu_ow - 1):
            spts_int_bybkg     =  np.concatenate((spts_int_bybkg, results[j].spts_int_bybkg), axis=1)
            bkg                =  np.concatenate((bkg, results[j].bkg), axis=1)
            bkg_variab         =  np.concatenate((bkg_variab, results[j].bkg_variab), axis=1)
            av_spts_int_bybkg  =  np.concatenate((av_spts_int_bybkg, results[j].av_spts_int_bybkg), axis=1)

        del results

        spts_int_bybkg[np.isnan(spts_int_bybkg)]        =  0
        av_spts_int_bybkg[np.isnan(av_spts_int_bybkg)]  =  0

        self.bkg             =  bkg
        self.bkg_variab      =  bkg_variab
        self.spts_int_bybkg  =  spts_int_bybkg


        book    =  xlwt.Workbook(encoding='utf-8')                                              # write results in xls file
        sheet1  =  book.add_sheet("Spots by Background")
        sheet2  =  book.add_sheet("Average Background")
        sheet3  =  book.add_sheet("Background Std")
        sheet4  =  book.add_sheet("Average Spots by Background")

        sheet1.write(0, 0, "Folder Name")
        sheet1.write(0, 1, foldername)
        sheet1.write(1, 0, "date")
        sheet1.write(1, 1, datetime.datetime.now().strftime("%d-%b-%Y"))

        for k in range(spts_id.size):
            sheet1.write(0, 3 + k, "SptId_" + str(spts_id[k]))
            sheet2.write(0, 3 + k, "SptId_" + str(spts_id[k]))
            sheet3.write(0, 3 + k, "SptId_" + str(spts_id[k]))
            sheet4.write(0, 3 + k, "SptId_" + str(spts_id[k]))
            for t in range(steps):
                sheet1.write(t + 1, 3 + k, spts_int_bybkg[k, t])
                sheet2.write(t + 1, 3 + k, bkg[k, t, 0])
                sheet3.write(t + 1, 3 + k, bkg_variab[k, t])
                sheet4.write(t + 1, 3 + k, av_spts_int_bybkg[k, t])

        book.save(foldername + '/SpotsIntensityDividedByBackground.xls')



class WriteSptsIntsDividedByBkgUtility:
    """Extract info, organized for multiprocessing"""
    def __init__(self, input_args):

        steps              =  input_args[1].shape[0]
        spts_int_bybkg     =  np.zeros((input_args[4].size, steps))
        av_spts_int_bybkg  =  np.zeros((input_args[4].size, steps))
        bkg                =  np.zeros((input_args[4].size, steps, 2))
        bkg_variab         =  np.zeros((input_args[4].size, steps))

        for k in range(input_args[4].size):
            spt_ints                 =  (input_args[1] * (input_args[3] == input_args[4][k])).sum(2).sum(1)                                             # intensity trace for a single spot
            spts_vol                 =  (input_args[2] * (input_args[3] == input_args[4][k])).sum(2).sum(1)                                             # volume trace for a single spot
            bffff                    =  SpotsBackgroundEstimation.SpotsBackgroundEstimation(input_args[3], input_args[0], input_args[4][k])             # extract info
            bkg[k, :, :]             =  bffff.bkg_values
            bkg_variab[k, :]         =  bffff.bkg_variab
            spts_int_bybkg[k, :]     =  spt_ints / bkg[k, :, 0]
            av_spts_int_bybkg[k, :]  =  spts_int_bybkg[k, :] / spts_vol

        self.spts_int_bybkg     =  spts_int_bybkg
        self.bkg                =  bkg
        self.bkg_variab         =  bkg_variab
        self.av_spts_int_bybkg  =  av_spts_int_bybkg


