"""This function removes the background from spots intensity.

Given a folder with analyzed data, the algorithm evaluates the
background value of each spot in each frame (using cages) and
subtracte it from the intensity of the spot. The background value
is multiplied by the volume of the spots of course. results
written in a .xls file.
"""

import datetime
import multiprocessing
import numpy as np
import xlwt

import AnalysisLoader
import SpotsBackgroundEstimation


class WriteSptsIntsMinusBkg:
    """Load the analysis, coordinate the multiprocessing and write results in .xls file"""
    def __init__(self, foldername):

        green4D           =  AnalysisLoader.RawData(foldername).green4D                                                     # load raw data spots 
        spots_3D          =  AnalysisLoader.SpotsIntsVol(foldername)                                                        # load spots volume info
        spots_tracked_3D  =  np.fromfile(foldername + '/spots_tracked.bin', 'uint16')                                       # load tracked spots
        spots_tracked_3D  =  spots_tracked_3D[3:].reshape((spots_tracked_3D[2], spots_tracked_3D[1], spots_tracked_3D[0]))
        spts_id           =  np.unique(spots_tracked_3D)[1:]                                                                # list of spots tags 
        steps             =  spots_tracked_3D.shape[0]                                                                      # number of time steps

        cpu_ow     =  multiprocessing.cpu_count()
        jobs_args  =  []
        t_chops    =  np.linspace(0, steps, cpu_ow).astype(np.int)
        for k in range(cpu_ow - 1):
            jobs_args.append([green4D[t_chops[k]:t_chops[k + 1], :, :, :], spots_3D.spots_ints[t_chops[k]:t_chops[k + 1], :, :], spots_3D.spots_vol[t_chops[k]:t_chops[k + 1], :, :], spots_tracked_3D[t_chops[k]:t_chops[k + 1], :, :], spts_id])     # prepare list of input for the multiprocessing

        pool     =  multiprocessing.Pool()
        results  =  pool.map(WriteSptsIntsMinusBkgUtility, jobs_args)
        pool.close()

        spts_int_nobkg  =  results[0].spts_int_nobkg                        # read and organized multiprocessing results
        bkg             =  results[0].bkg
        bkg_variab      =  results[0].bkg_variab
        for j in range(1, cpu_ow - 1):
            spts_int_nobkg  =  np.concatenate((spts_int_nobkg, results[j].spts_int_nobkg), axis=1)
            bkg             =  np.concatenate((bkg, results[j].bkg), axis=1)
            bkg_variab      =  np.concatenate((bkg_variab, results[j].bkg_variab), axis=1)

        del results

        self.bkg             =  bkg
        self.bkg_variab      =  bkg_variab
        self.spts_int_nobkg  =  spts_int_nobkg


        book    =  xlwt.Workbook(encoding='utf-8')                            # wrrite results in xls file
        sheet1  =  book.add_sheet("Sheet 1")
        sheet2  =  book.add_sheet("Average Background")
        sheet3  =  book.add_sheet("Background Std")

        sheet1.write(0, 0, "Folder Name")
        sheet1.write(0, 1, foldername)
        sheet1.write(1, 0, "date")
        sheet1.write(1, 1, str(datetime.date.today()))

        for k in range(spts_id.size):
            sheet1.write(0, 3 + k, "SptId_" + str(spts_id[k]))
            for t in range(steps):
                sheet1.write(t + 1, 3 + k, spts_int_nobkg[k, t])
                sheet2.write(t + 1, 3 + k, bkg[k, t, 0])
                sheet3.write(t + 1, 3 + k, bkg_variab[k, t])


        book.save(foldername + '/SpotsIntensityMinusBackground.xls')



class WriteSptsIntsMinusBkgUtility:
    """Utility function for multiprocessing"""
    def __init__(self, input_args):

        steps           =  input_args[1].shape[0]
        spts_int_nobkg  =  np.zeros((input_args[4].size, steps))
        bkg             =  np.zeros((input_args[4].size, steps, 2))
        bkg_variab      =  np.zeros((input_args[4].size, steps))
        for k in range(input_args[4].size):
            spt_ints              =  (input_args[1] * (input_args[3] == input_args[4][k])).sum(2).sum(1)                                        # intensity of the spot
            bffff                 =  SpotsBackgroundEstimation.SpotsBackgroundEstimation(input_args[3], input_args[0], input_args[4][k])        # measure background value
            bkg[k, :, :]          =  bffff.bkg_values
            bkg_variab[k, :]      =  bffff.bkg_variab
            spt_vol               =  (input_args[2] * (input_args[3] == input_args[4][k])).sum(2).sum(1)                                        # volume of the spot
            spts_int_nobkg[k, :]  =  spt_ints - bkg[k, :, 0] * spt_vol                                                                          # intensity minus background for the spot

        self.spts_int_nobkg  =  spts_int_nobkg
        self.bkg             =  bkg
        self.bkg_variab      =  bkg_variab


